
#
# Mongrel2-Transceiver -- Communication Protocol Python Parser and Originator
#
# Copyright (c) 2015, Web Heroes Inc..
#
# Mongrel2-Transceiver is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later version.  See
# the LICENSE file at the top of the source tree.
#
# Mongrel2-Transceiver is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#

from transceiver	import *
from server		import *

__author__                      = "Matthew Brisebois"
__email__                       = "matthew@webheroes.ca"
__copyright__                   = "Copyright (c) 2015 Web Heroes Inc."
__license__                     = "Dual License: GPLv3 (or later) and Commercial (see LICENSE)"

__all__			= ["Transceiver","Server","Connector","Client"]
